from .compat import json


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        from .api import StripeObject
        
        if isinstance(obj, StripeObject):
            return obj.to_dict()
        return super(JSONEncoder, self).default(obj)


# -*- coding: utf-8 -*-
"""
Stripe Python bindings
~~~~~~~~~~~~~~~~~~~~~~

stripe-requests is a replacement for stripe-python using only the requests 
library. Basic usage:

   >>> from stripe import Stripe
   >>> stripe = Stripe(api_key='MY_API_KEY')
   >>> stripe.Account.retrieve()

Full documentation is at <https://stripe-requests.readthedocs.org>.

:copyright: (c) 2013 by Allan Lei.
:license: MIT, see LICENSE for more details.

"""

__title__ = 'stripe-requests'
__version__ = '1.9.1-dev'
__author__ = 'Allan Lei'
__license__ = 'MIT'
__copyright__ = 'Copyright 2013 Allan Lei'


import os
import platform
import logging
import requests


from .compat import *
from .transport import OAuth2Auth, HTTPAdapter
from .encoders import JSONEncoder
from .errors import *
from .api import *

__all__ = ['StripeObject', 'Charge', 'Coupon', 'Plan', 'Transfer', 'Recipient', 
    'JSONEncoder', 'StripeObjectEncoder', 'StripeError', 'APIError', 'APIConnectionError', 
    'InvalidRequestError', 'AuthenticationError', 'CardError', 'stripe', 'Stripe']

StripeObjectEncoder = JSONEncoder

STRIPE_API_BASE_V1 = 'https://api.stripe.com/v1/'
STRIPE_API_BASE = STRIPE_API_BASE_V1

logger = logging.getLogger('stripe')


class Stripe(object):
    def __init__(self, api_key=None, verify_ssl_certs=True, api_version=None, 
            api_base=STRIPE_API_BASE, session_class=requests.Session):
        super(Stripe, self).__init__()
        self.session = session_class()
        self.session.mount('http://', HTTPAdapter())
        self.session.mount('https://', HTTPAdapter())
        self.session.auth = OAuth2Auth()
        self.api_key = api_key
        self.verify_ssl_certs = verify_ssl_certs
        self.api_version = api_version
        self.api_base = api_base

        self.session.headers.update({
            'User-Agent' : 'Stripe/v1 PythonBindings/{version}'.format(version=__version__),
            'X-Stripe-Client-User-Agent' : json.dumps({
                'bindings_version' : __version__,
                'lang': 'python',
                'implementation': platform.python_implementation(),
                'publisher' : 'stripe',
                'httplib': requests.__title__,
                'lang_version': platform.python_version(),
                'platform': platform.platform(),
                'uname': ' '.join(platform.uname()),
            }),
        })

        # TODO: This is weird, mainly to keep the interface the same as stripe-python
        for res in [
            'Account',
            'Charge',
            'Customer',
            'Invoice',
            'InvoiceItem',
            'Plan',
            'Token',
            'Coupon',
            'Event',
            'Transfer',
            'Recipient']:
            Resource = type(res, (globals()[res], ), {
                'stripe': self,
            })
            setattr(self, res, Resource)

    def request(self, method, url, *args, **kwargs):
        if 'api_key' in kwargs:
            kwargs.setdefault('auth', OAuth2Auth(kwargs.pop('api_key')))
        url = urljoin(self.api_base, url)

        data = kwargs.get('data', None)
        if data and not isinstance(data, basestring):
            kwargs['data'] = urlencode(data)

        params = kwargs.get('params', None)
        if params and not all([isinstance(value, (basestring, int)) for value in params.values()]):
            url_parts = list(urlparse(url))
            query = dict(parse_qsl(url_parts[4]))
            query.update(kwargs.pop('params'))
            url_parts[4] = urlencode(query)
            url = urlunparse(url_parts)

        try:
            response = self.session.request(method, url, *args, **kwargs)
        except requests.exceptions.ConnectionError as e:
            raise InsecureConnectionError(str(e))
        except requests.exceptions.RequestException as e:
            # Reraise as APIConnectionError
            raise APIConnectionError(str(e))

        response.raise_for_status()
        return response.json()

    @property
    def api_key(self):
        return self.session.auth.api_key

    @api_key.setter
    def api_key(self, key):
        self.session.auth.api_key = key

    @property
    def verify_ssl_certs(self):
        return bool(self.session.verify)

    @verify_ssl_certs.setter
    def verify_ssl_certs(self, verify):
        self.session.verify = bool(verify)

    @property
    def api_version(self):
        return self.session.headers.get('Stripe-Version', None)

    @api_version.setter
    def api_version(self, version):
        if not version:
            return
        self.session.headers.update({
            'Stripe-Version': version,
        })


# A default Stripe instance, api_key from env var
stripe = Stripe(api_key=os.environ.get('STRIPE_API_KEY'))


from requests.compat import (basestring, str, urlparse, urlunparse, json, 
    quote_plus, urljoin, is_py2, is_py3)
from requests.compat import urlencode as urllib_urlencode

if is_py2:
    from urlparse import parse_qsl
elif is_py3:
    from urllib.parse import parse_qsl

import time
import datetime
import sys



def utf8(value):
    if isinstance(value, str) and sys.version_info < (3, 0):
        return value.encode('utf-8')
    return value


def urlencode(obj):
    def urlencode_list(stk, key, listvalue):
        for v in listvalue:
            v = utf8(v)
            stk.append(("%s[]" % (key), v))

    def urlencode_datetime(stk, key, dttime):
        utc_timestamp = int(time.mktime(dttime.timetuple()))
        stk.append((key, utc_timestamp))

    def urlencode_none(stk, k, v):
        pass

    def urlencode_dict(stk, key, dictvalue):
        n = {}
        for k, v in dictvalue.items():
            k, v = utf8(k), utf8(v)
            n["%s[%s]" % (key, k)] = v
        stk.extend(urlencode_inner(n))

    def urlencode_inner(obj):
        ENCODERS = {
            list: urlencode_list,
            dict: urlencode_dict,
            datetime.datetime: urlencode_datetime,
            None.__class__: urlencode_none,
        }

        stk = []
        for key, value in obj.items():
            key = utf8(key)
            try:
                encoder = ENCODERS[value.__class__]
                encoder(stk, key, value)
            except KeyError:
                # don't need special encoding
                value = utf8(value)
                stk.append((key, value))
        return stk
    return urllib_urlencode(urlencode_inner(obj))


# -*- coding: utf-8 -*-
import datetime
import string
import random
import unittest

from stripe.compat import json


# dummy information used in the tests below
NOW = datetime.datetime.now()

DUMMY_CARD = {
    'number': '4242424242424242',
    'exp_month': NOW.month,
    'exp_year': NOW.year + 4
}
DUMMY_CHARGE = {
    'amount': 100,
    'currency': 'usd',
    'card': DUMMY_CARD
}

DUMMY_PLAN = {
    'amount': 2000,
    'interval': 'month',
    'name': 'Amazing Gold Plan',
    'currency': 'usd',
    'id': 'stripe-test-gold-' + ''.join(random.choice(string.ascii_lowercase) for x in range(10))
}

DUMMY_COUPON = {
    'percent_off': 25,
    'duration': 'repeating',
    'duration_in_months': 5
}

SAMPLE_INVOICE = json.loads("""
{
  "amount_due": 1305,
  "attempt_count": 0,
  "attempted": true,
  "charge": "ch_wajkQ5aDTzFs5v",
  "closed": true,
  "customer": "cus_osllUe2f1BzrRT",
  "date": 1338238728,
  "discount": null,
  "ending_balance": 0,
  "id": "in_t9mHb2hpK7mml1",
  "livemode": false,
  "next_payment_attempt": null,
  "object": "invoice",
  "paid": true,
  "period_end": 1338238728,
  "period_start": 1338238716,
  "starting_balance": -8695,
  "subtotal": 10000,
  "total": 10000,
  "lines": {
    "invoiceitems": [],
    "prorations": [],
    "subscriptions": [
      {
        "plan": {
          "interval": "month",
          "object": "plan",
          "identifier": "expensive",
          "currency": "usd",
          "livemode": false,
          "amount": 10000,
          "name": "Expensive Plan",
          "trial_period_days": null,
          "id": "expensive"
        },
        "period": {
          "end": 1340917128,
          "start": 1338238728
        },
        "amount": 10000
      }
    ]
  }
}
""")

if __name__ == '__main__':
    from .test_stripe import *
    unittest.main()


import warnings
import requests

from requests.structures import CaseInsensitiveDict
from requests.utils import get_encoding_from_headers
from requests.cookies import extract_cookies_to_jar

from . import errors



class OAuth2Auth(requests.auth.AuthBase):
    def __init__(self, api_key=None, required=True, restrict_to_https=True):
        self.api_key = api_key
        self.required = required
        self.restrict_to_https = restrict_to_https

    def __call__(self, request):
        if self.required and not self.api_key:
            raise errors.AuthenticationError('No API key provided.')

        if not request.url.startswith('https://'):
            warnings.warn('Sending API key over unsecure connection')
            
            if self.restrict_to_https:
                raise errors.InsecureConnectionError()

        request.headers['Authorization'] = 'Bearer {api_key}'.format(
            api_key=self.api_key)
        return request


class Response(requests.models.Response):
    def raise_for_status(self):
        HTTPErrorClass = None
        http_error_msg = None
        error = {}

        try:
            error = self.json()['error']
        except KeyError:
            pass

        if self.status_code in [requests.codes.bad_request, requests.codes.not_found]:
            HTTPErrorClass, http_error_msg = errors.InvalidRequestError, error.get('message')
        elif self.status_code in [requests.codes.unauthorized]:
            HTTPErrorClass, http_error_msg = errors.AuthenticationError, error.get('message')
        elif self.status_code in [requests.codes.payment_required]:
            HTTPErrorClass, http_error_msg = errors.CardError, error.get('message')
        elif self.status_code in [requests.codes.forbidden]:
            if not self.request.url.startswith('https://'):
                HTTPErrorClass, http_error_msg = errors.InsecureConnectionError, 'Insecure connection'

        if http_error_msg and HTTPErrorClass:
            raise HTTPErrorClass(http_error_msg, response=self)
        return super(Response, self).raise_for_status()


class HTTPAdapter(requests.adapters.HTTPAdapter):
    def build_response(self, req, resp):
        """
        Copy of requests.adapters.HTTPAdapter(1.9.1). 
        Overriding to use a custom Response class
        """
        response = Response()

        # Fallback to None if there's no status_code, for whatever reason.
        response.status_code = getattr(resp, 'status', None)

        # Make headers case-insensitive.
        response.headers = CaseInsensitiveDict(getattr(resp, 'headers', {}))

        # Set encoding.
        response.encoding = get_encoding_from_headers(response.headers)
        response.raw = resp
        response.reason = response.raw.reason

        if isinstance(req.url, bytes):
            response.url = req.url.decode('utf-8')
        else:
            response.url = req.url

        # Add new cookies from the server.
        extract_cookies_to_jar(response.cookies, req, resp)

        # Give the Response some context.
        response.request = req
        response.connection = self
        return response


import requests


class StripeError(requests.exceptions.HTTPError):
    @property
    def http_body(self):
        return self.response.text

    @property
    def http_status(self):
        return self.response.status_code

    @property
    def json_body(self):
        return self.response.json()

class APIError(StripeError):
    pass

class APIConnectionError(StripeError):
    pass

class InsecureConnectionError(APIConnectionError):
    pass

class InvalidRequestError(StripeError):
    @property
    def param(self):
        return self.response.request.params

class AuthenticationError(StripeError):
    pass

class CardError(StripeError):
    @property
    def param(self):
        return self.response.request.params

    @property
    def code(self):
        return self.response.json().get('error', {}).get('code')


import logging

from .base import *

logger = logging.getLogger('stripe')



class Account(SingletonAPIResource):
    resource_uri = 'account'

class Charge(CreateableAPIResource, ListableAPIResource, UpdateableAPIResource):
    resource_uri = 'charges'

    def refund(self, **params):
        requestor = self.stripe
        response = requestor.request('post', self.resource_uri + '/refund', data=params)
        self.refresh_from(response)
        return self

    def capture(self, **params):
        requestor = self.stripe
        response = requestor.request('post', self.resource_uri + '/capture', data=params)
        self.refresh_from(response)
        return self

    def update_dispute(self, **params):
        requestor = self.stripe
        response = requestor.request('post', self.resource_uri + '/dispute', data=params)
        self.refresh_from({ 'dispute' : response }, True)
        return self.dispute

class Customer(CreateableAPIResource, UpdateableAPIResource, ListableAPIResource, DeletableAPIResource):
    resource_uri = 'customers'

    def add_invoice_item(self, **params):
        params['customer'] = self.id
        ii = InvoiceItem.create(**params)
        return ii

    def invoices(self, **params):
        params['customer'] = self.id
        invoices = Invoice.all(stripe=self.stripe, **params)
        return invoices

    def invoice_items(self, **params):
        params['customer'] = self.id
        iis = InvoiceItem.all(stripe=self.stripe, **params)
        return iis

    def charges(self, **params):
        params['customer'] = self.id
        charges = Charge.all(stripe=self.stripe, **params)
        return charges

    def update_subscription(self, **params):
        requestor = self.stripe
        response = requestor.request('post', self.resource_uri + '/subscription', data=params)
        self.refresh_from({ 'subscription' : response }, True)
        return self.subscription

    def cancel_subscription(self, **params):
        requestor = self.stripe
        response = requestor.request('delete', self.resource_uri + '/subscription', data=params)
        self.refresh_from({ 'subscription' : response }, True)
        return self.subscription

    def delete_discount(self, **params):
        requestor = self.stripe
        response = requestor.request('delete', self.resource_uri + '/discount')
        self.refresh_from({ 'discount' : None }, True)

class Invoice(CreateableAPIResource, ListableAPIResource, UpdateableAPIResource):
    resource_uri = 'invoices'

    def pay(self):
        from .utils import convert_to_stripe_object
        requestor = self.stripe
        response, = requestor.request('post', self.resource_uri + '/pay', data={})
        return convert_to_stripe_object(response, stripe=stripe)

    @classmethod
    def upcoming(cls, stripe=None, **params):
        stripe = stripe or cls.stripe
        from .utils import convert_to_stripe_object
        requestor = stripe
        response = requestor.request('get', cls.resource_uri + '/upcoming', params=params)
        return convert_to_stripe_object(response, stripe=stripe)

class InvoiceItem(CreateableAPIResource, UpdateableAPIResource, ListableAPIResource, DeletableAPIResource):
    resource_uri = 'invoiceitems'

class Plan(CreateableAPIResource, DeletableAPIResource, UpdateableAPIResource, ListableAPIResource):
    resource_uri = 'plans'

class Token(CreateableAPIResource):
    resource_uri = 'tokens'

class Coupon(CreateableAPIResource, DeletableAPIResource, ListableAPIResource):
    resource_uri = 'coupons'

class Event(ListableAPIResource):
    resource_uri = 'events'

class Transfer(CreateableAPIResource, ListableAPIResource):
    resource_uri = 'transfers'

class Recipient(CreateableAPIResource, UpdateableAPIResource, ListableAPIResource, DeletableAPIResource):
    resource_uri = 'recipients'

    def transfers(self, **params):
        params['recipient'] = self.id
        transfers = Transfer.all(stripe=self.stripe, **params)
        return transfers


from .compat import *


def convert_to_stripe_object(resp, stripe=None):
    """
    Initializes a StripeObject from a JSON response
    """
    from .api import (Charge, Customer, Invoice, InvoiceItem, Plan, 
        Coupon, Token, Event, Transfer, ListObject, Recipient, StripeObject)

    types = {
        'charge': Charge, 
        'customer': Customer,
        'invoice' : Invoice, 
        'invoiceitem' : InvoiceItem,
        'plan': Plan, 
        'coupon': Coupon, 
        'token': Token, 
        'event': Event,
        'transfer': Transfer, 
        'list': ListObject, 
        'recipient': Recipient,
    }

    if isinstance(resp, list):
        return [convert_to_stripe_object(i, stripe=stripe) for i in resp]
    elif isinstance(resp, dict):
        resp = resp.copy()
        klass_name = resp.get('object')
        if isinstance(klass_name, basestring):
            klass = types.get(klass_name, StripeObject)
        else:
            klass = StripeObject
        klass.stripe = stripe
        return klass.construct_from(resp)
    else:
        return resp


import logging

from ..errors import InvalidRequestError
from ..utils import utf8, convert_to_stripe_object
from ..encoders import JSONEncoder
from ..compat import *

logger = logging.getLogger('stripe')



class StripeObject(object):
    _permanent_attributes = set(['stripe', 'resource_uri'])
    stripe = None
    resource_uri = None

    def __init__(self, id=None, stripe=None, **params):
        self.__dict__['_values'] = set()
        self.__dict__['_unsaved_values'] = set()
        self.__dict__['_transient_values'] = set()
        self.__dict__['_retrieve_params'] = params
        self.stripe = stripe or self.__class__.stripe

        if id:
            self.id = id
            self.resource_uri = '{base}/{id}'.format(
                base=self.__class__.resource_uri,
                id=quote_plus(utf8(id)))
        else:
            self.resource_uri = None

    def __setattr__(self, k, v):
        self.__dict__[k] = v
        self._values.add(k)
        if k not in self._permanent_attributes:
            self._unsaved_values.add(k)

    def __getattr__(self, k):
        try:
            return self.__dict__[k]
        except KeyError:
            pass
        if k in self._transient_values:
            raise AttributeError("%r object has no attribute %r.  HINT: The %r attribute was set in the past, however.  It was then wiped when refreshing the object with the result returned by Stripe's API, probably as a result of a save().  The attributes currently available on this object are: %s" %
                    (type(self).__name__, k, k, ', '.join(self._values)))
        else:
            raise AttributeError("%r object has no attribute %r" % (type(self).__name__, k))

    def __getitem__(self, k):
        if k in self._values:
            return self.__dict__[k]
        elif k in self._transient_values:
            raise KeyError("%r.  HINT: The %r attribute was set in the past, however.  It was then wiped when refreshing the object with the result returned by Stripe's API, probably as a result of a save().  The attributes currently available on this object are: %s" % (k, k, ', '.join(self._values)))
        else:
            raise KeyError(k)

    def get(self, k, default=None):
        try:
            return self[k]
        except KeyError:
            return default

    def setdefault(self, k, default=None):
        try:
            return self[k]
        except KeyError:
            self[k] = default
        return default

    def __setitem__(self, k, v):
        setattr(self, k, v)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    @classmethod
    def construct_from(cls, values, stripe=None):
        stripe = stripe or cls.stripe
        instance = cls(values.get('id'), stripe=stripe)
        instance.refresh_from(values)
        return instance

    def refresh_from(self, values, partial=False):
        # Wipe old state before setting new.  This is useful for e.g. updating a
        # customer, where there is no persistent card parameter.  Mark those values
        # which don't persist as transient
        if partial:
            removed = set()
        else:
          removed = self._values - set(values)

        for k in removed:
            if k in self._permanent_attributes:
                continue
            del self.__dict__[k]
            self._values.discard(k)
            self._transient_values.add(k)
            self._unsaved_values.discard(k)

        for k, v in values.items():
            if k in self._permanent_attributes:
                continue
            self.__dict__[k] = convert_to_stripe_object(v, stripe=self.stripe)
            self._values.add(k)
            self._transient_values.discard(k)
            self._unsaved_values.discard(k)

    def __repr__(self):
        type_string = ''
        if isinstance(self.get('object'), basestring):
            type_string = ' %s' % self.get('object').encode('utf8')

        id_string = ''
        if isinstance(self.get('id'), basestring):
            id_string = ' id=%s' % self.get('id').encode('utf8')

        return '<%s%s%s at %s> JSON: %s' % (type(self).__name__, type_string, id_string, hex(id(self)), json.dumps(self.to_dict(), sort_keys=True, indent=2, cls=JSONEncoder))

    def __unicode__(self):
        return json.dumps(self.to_dict(), sort_keys=True, indent=2, cls=JSONEncoder)

    def to_dict(self):
        def _serialize(o):
            if isinstance(o, StripeObject):
                return o.to_dict()
            if isinstance(o, list):
                return [_serialize(i) for i in o]
            return o

        d = dict()
        for k in sorted(self._values):
            if k in self._permanent_attributes:
                continue
            v = getattr(self, k)
            v = _serialize(v)
            d[k] = v
        return d




class APIResource(StripeObject):
    def _ident(self):
        return [self.get('id')]

    @classmethod
    def retrieve(cls, id, stripe=None, **params):
        stripe = stripe or cls.stripe
        instance = cls(id, stripe=stripe, **params)
        instance.refresh()
        return instance

    def refresh(self):
        requestor = self.stripe

        if not self.resource_uri:
            id = self.get('id')
            raise InvalidRequestError('Could not determine which URL to request: %s instance has invalid ID: %r' % (type(self).__name__, id), 'id')
        response = requestor.request('get', self.resource_uri, params=self._retrieve_params)
        self.refresh_from(response)
        return self


class ListObject(StripeObject):
    def all(self, **params):
        requestor = self.stripe
        response = requestor.request('get', self.resource_uri, params=params)
        return convert_to_stripe_object(response, stripe=self.stripe)

class SingletonAPIResource(APIResource):
    def __init__(self, *args, **kwargs):
        super(SingletonAPIResource, self).__init__(*args, **kwargs)
        self.resource_uri = self.__class__.resource_uri

    def _ident(self):
        return [self.get('id')]

    @classmethod
    def retrieve(cls, stripe=None):
        stripe = stripe or cls.stripe
        instance = cls(None, stripe=stripe)
        instance.refresh()
        return instance




# Classes of API operations
class ListableAPIResource(APIResource):
    @classmethod
    def all(cls, stripe=None, **params):
        stripe = stripe or cls.stripe
        requestor = stripe
        response = requestor.request('get', cls.resource_uri, params=params)
        list_obj = convert_to_stripe_object(response, stripe=stripe)
        list_obj.resource_uri = cls.resource_uri
        return list_obj

class CreateableAPIResource(APIResource):
    @classmethod
    def create(cls, stripe=None, **params):
        stripe = stripe or cls.stripe
        requestor = stripe
        response = requestor.request('post', cls.resource_uri, data=params)
        return convert_to_stripe_object(response, stripe=stripe)

class UpdateableAPIResource(APIResource):
    def save(self):
        if self._unsaved_values:
            requestor = self.stripe

            params = {}
            for k in self._unsaved_values:
                if k == 'id':
                    continue
                params[k] = getattr(self, k)

            response = requestor.request('post', self.resource_uri, data=params)
            self.refresh_from(response)
        else:
            logger.debug("Trying to save already saved object %r" % (self, ))
        return self

class DeletableAPIResource(APIResource):
    def delete(self, **params):
        requestor = self.stripe
        response = requestor.request('delete', self.resource_uri, params=params)
        self.refresh_from(response)
        return self


import os
import sys
import multiprocessing      # Weird nosetests fix

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

from stripe import __version__

setup(
    name='stripe-requests',
    version=__version__,
    description='Stripe python bindings using requests',
    author='Allan Lei',
    author_email='allanlei@helveticode.com',
    url='https://github.com/allanlei/stripe-requests',
    license=open('LICENSE').read(),
    packages=find_packages(),
    install_requires=[
        'requests >= 1.2.0, < 1.3.0',
    ],
    tests_require=[
        'nose',
    ],
    test_suite = 'nose.collector',
    classifiers=(
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: PyPy',
    ),
)